import re
from collections import OrderedDict
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import indent
from typing import Any, Dict, List

import yaml
from pydantic import BaseModel, Field, ValidationInfo, field_validator, model_validator

from botocraft.sync.docstring import DocumentationFormatter

DATA_DIR = Path(__file__).parent.parent / "data"
SERVICES_DIR = Path(__file__).parent.parent / "services"
DOCS_DIR = Path(__file__).parent.parent.parent / "doc" / "source"


# ------
# Models
# ------


class Importable(BaseModel):
    """
    A configuration object for a mixin class to add to a model.
    """

    #: The name of the mixin class
    name: str
    #: The botocraft import path
    import_path: str


class RegexTransformer(BaseModel):
    """
    Used when generating a property on a model that is a regular expression
    transformation of an attribute on the model.

    If the regular expression has named groups, we'll return a dictionary
    of the named groups.  Otherwise, we'll return the value of the first
    group.
    """

    #: The name of the attribute on our model to use as
    #: the input to the regular expression.
    attribute: str
    #: Use this regular expression to transform the attribute value.
    regex: str

    @field_validator("regex")
    @classmethod
    def check_valid_regexp(
        cls,
        v: str | None,
        info: ValidationInfo,
    ) -> str | None:
        """
        Validate that the transformer is a valid regular expression.

        Args:
            v: the value of field
            info: pydantic field validation info

        Raises:
            ValueError: if the transformer is not a valid regular expression

        Returns:
            The value of the field

        """
        if v is not None:
            try:
                re.compile(v)
            except re.error as exc:
                msg = f"{info.field_name} must be a valid regular expression"
                raise ValueError(msg) from exc
        return v


class CodeTransformerDefinition(BaseModel):
    """
    Used when generating a property on a model that is a verbatim expression
    to return.
    """

    #: The name of the attribute on our model to use as
    #: the input to the regular expression.
    code: str
    #: Use this code to transform the attribute value.
    return_type: str


class AttributeTransformerDefinition(BaseModel):
    #: If specified, use this regular expression to build the
    #: property output
    regex: RegexTransformer | None = None
    #: If specified, use this mapping of attribute values to
    #: output keys to generate the output
    mapping: Dict[str, str] | None = None
    #: If specified, use this property as an alias for this
    #: attribute.  This may be an alias on another related
    #: model
    alias: str | None = None
    #: If specified, use this property as a verbatim expression
    #: to return
    #: attribute.  This may be an alias on another related
    #: model
    code: CodeTransformerDefinition | None = None

    @model_validator(mode="before")
    @classmethod
    def check_transformer(cls, data: Any) -> Any:
        """
        Validate that only one of a regex, a mapping or an alias is specified,
        but not more than one.

        Args:
            data: the input data for this model

        Raises:
            ValueError: both a regex and a mapping are specified

        Returns:
            The input data for this model

        """
        transformers = [
            data.get("regex"),
            data.get("mapping"),
            data.get("alias"),
            data.get("code"),
        ]
        if transformers.count(None) < 2:  # noqa: PLR2004
            # More than one is specified
            msg = "Only one of regex, mapping, alias or code can be specified"
            raise ValueError(msg)
        if transformers.count(None) == 4:  # noqa: PLR2004
            # Nothing is specified
            msg = "One of regex, mapping, alias  or code must be specified"
            raise ValueError(msg)
        return data


class ModelPropertyDefinition(BaseModel):
    """
    The definition of a single property on a :py:class:`ModelDefinition`.

    One of ``regex``, ``mapping`` or ``alias`` must be specified.

    If :py:attr:`cached` is ``True``, we'll cache the value of this property
    using :py:func:`functools.cached_property`.
    """

    #: The docstring for this property
    docstring: str | None = None
    #: If ``True``, make this property be cached
    cached: bool = False
    #: If specified, use this definition to build the property
    #: output from the model fields
    transformer: AttributeTransformerDefinition


class ModelRelationshipDefinition(BaseModel):
    """
    Defines a property on a model that is a relationship to another model.
    """

    #: The docstring for this property
    docstring: str | None = None
    #: If ``True``, make this property be cached
    cached: bool = True
    #: If specified, use this definition to build the primary key
    #: for the related model from data from this model
    #:
    #: .. important::
    #:
    #:     If you are using a regex transformer, you must use named groups.
    #:
    #:     Example::
    #:         r'arn:aws:ecs:(?P<region>[^:]+):(?P<account_id>[^:]+):(?P<resource_type>[^:]+)/(?P<resource_id>[^:]+)'  # noqa: E501
    transformer: AttributeTransformerDefinition
    #: The name of the botocraft primary model to use for this relationship
    #:
    #: .. important::
    #:
    #:     This needs to be the name of a primary model, not a secondary model.
    #:     Only primary models have managers.
    primary_model_name: str
    #: If ``True``, force this to be a many-to-many relationship.  Otherwise,
    #: try to figure it out from our transformer.
    many: bool | None = None
    #: If specified, use this as the name of the method on the related model's
    #: manager to get the related model instance.  This defaults to ``get``
    #: if :py:attr:`many` is ``False`` and ``list`` if :py:attr:`many` is ``True``.
    method: str | None = None

    @model_validator(mode="before")
    @classmethod
    def check_regex_transformer(cls, data: Any) -> Any:
        """
        If we have a regex transformer, make sure that it only uses
        named groups.

        Args:
            data: the input data for this model

        Raises:
            ValueError: if the regex transformer has no named groups
                or has unnamed groups

        Returns:
            The input data for this model

        """
        transformer = data["transformer"]
        regex_transformer = transformer.get("regex")
        if regex_transformer is not None:
            pattern = regex_transformer["regex"]
            regex = re.compile(rf"{pattern}")
            if not regex.groupindex:
                msg = "Regex transformer must have named groups"
                raise ValueError(msg)
            if len(regex.groupindex.keys()) != regex.groups:
                msg = "Regex transformer must have no unnamed groups"
                raise ValueError(msg)
        return data

    @model_validator(mode="before")
    @classmethod
    def check_alias_transformer(cls, data: Any) -> Any:
        """
        Don't allow alias transformers for relations -- we need to know
        the argument name for the related model, and alias types don't
        support that.

        Args:
            data: the input data for this model

        Raises:
            ValueError: an alias transformer is specified

        Returns:
            The input data for this model

        """
        if "transformer" not in data:
            msg = f"Transformer is required for ModelRelationshipDefinition: {data}"
            raise ValueError(msg)
        if data["transformer"].get("alias") is not None:
            msg = (
                "Alias transformers are not supported for relations. Use a regex "
                "or mapping transformer instead."
            )
            raise ValueError(msg)
        return data


class ModelManagerMethodArgDefinition(BaseModel):
    #: The name of the argument
    name: str
    #: The value of the argument.  If ``None``, we'll use the value of the
    #: attribute on the model with the same name.
    value: str | None = None
    #: The type of the argument
    attr_type: str | None = None


class ModelManagerMethodKwargDefinition(BaseModel):
    #: The name of the argument
    name: str
    #: The value of the argument.  if ``None``, we'll use the value of the
    #: attribute on the model with the same name.
    value: str | None = None
    #: The type of the argument
    attr_type: str | None = None
    #: The default value for the argument
    default: Any | None = None


class ModelManagerMethodDefinition(BaseModel):
    """
    Used to provide methods on a model that call methods on the manager for the
    model.
    """

    # The name of the method on the manager
    manager_method: str
    #: If ``True``, make this result from this cached
    cached: bool = False
    #: The docstring for the method
    docstring: str | None = None
    #: The arguments for the method.  These should refer to attributes : on the
    #: model and map them to args on the manager method in the proper : order.
    #: The key is the position of the argument in the manager method signature,
    #: value is the name of the attribute on the model.
    args: Dict[int, ModelManagerMethodArgDefinition] = Field(default_factory=dict)
    #: Args the users must supply.  The key is the position of the argument in
    #: the manager method signature, and the value is the name of the argument
    #: to add to the model method signature.
    user_args: Dict[int, ModelManagerMethodArgDefinition] = Field(default_factory=dict)
    #: The keyword arguments for the method.  These should refer to attributes
    #: on the model and map them to kwargs on the manager method with the proper
    #: keyword.
    keyword_args: List[ModelManagerMethodKwargDefinition] = Field(default_factory=list)
    #: Keyword arguments the users can supply
    user_keyword_args: List[ModelManagerMethodKwargDefinition] = Field(
        default_factory=list
    )
    #: A list of decorators to wrap the method in
    decorators: List[Importable] = Field(default_factory=list)


class ModelAttributeDefinition(BaseModel):
    """
    The definition of a single attribute on a :py:class:`ModelDefinition`.

    We use this to override the default configuration of a model attribute
    that we generate from the botocore model.
    """

    #: If specified, use this as the name of the field in the model.
    #: Set the original name of the field as a pydantic field alias.
    rename: str | None = None
    #: If ``True``, make this field immutable.
    readonly: bool = False
    #: If ``True``, force the field to be required, even if the underlying
    #: botocore field is not required.
    required: bool = False
    #: If specified, set this as the default value for the field.
    default: Any | None = None
    #: If specified, use this as the docstring for the field.
    docstring: str | None = None
    #: If specified, use this as the python type for the field.
    python_type: str | None = None
    #: If specified, the list of imports to add to the top of the
    #: to support the python type.
    imports: List[str] = []
    #: This is not defined in the YAML file, but is set by
    #: :py:meth:`ModelGenerator.fields` during generation.
    botocore_shape: Any | None = None


class ModelDefinition(BaseModel):
    """
    The definition of a single model for a :py:class:`ServiceDefinition`.
    """

    #: The botocore model name
    name: str
    #: A list of mixin classes to add to the model
    mixins: List[Importable] = []
    #: If ``True``, make this model immutable.
    readonly: bool = False
    #: The plural form of our model name, if different from
    #: what the inflect library would generate.
    plural: str | None = None
    #: The name of the base class for this model.  If not specified, we'll
    #: use the default base class for the use case.
    base_class: str | None = None
    #: A different docstring for the model.  If not specified, we'll use the
    #: docstring from the botocore model.
    docstring: str | None = None
    #: If defined, use this as the model name instead of what botocore supplies.
    #: We use this if some submodels have a ame conflict with a model in
    #: another service or an attribute on a model in this service.
    alternate_name: str | None = None
    #: The primary key for this model.  This is the field that will be used
    #: to determine whether we need to create a new model instance or update
    #: an existing one.  This can be the name of a property.
    primary_key: str | None = None
    #: The ``ARN`` key for this model attribute if any.  Some AWS models
    #: have an ARN, a name, and an ID.  Some have no ARN.
    arn_key: str | None = None
    #: The ``name`` key for this model attribute if any.  Some AWS models
    #: have an ARN, a name, and an ID.  Some have no name.
    name_key: str | None = None
    #: A list of field overrides for this model.  If a field name is not
    #: specified in this list, it will be generated verbatim from the
    #: botocore model attribute definition.
    fields: Dict[str, ModelAttributeDefinition] = {}
    #: A list of extra fields for this model.  These are fields that are
    #: not in the botocore model, but are either are needed at create time
    #: or because they turn into something else in the response.  This is
    #: pretty rarely used.
    extra_fields: Dict[str, ModelAttributeDefinition] = {}
    #: The name of the output shape for the get method for this model.  We use
    #: this to add to the model any fields that appear in the get method
    #: response, but which are not present in the create method response.
    output_shape: str | None = None
    #: The names of the request shapes for the writable methods for this model.
    #: We use these to determine which fields on the model are writable.
    input_shapes: List[str] | None = None
    #: Computed properties.  These are all defined by us in the botocraft
    #: config.
    properties: Dict[str, ModelPropertyDefinition] = {}
    #: Relationships to other models.  These are all defined by us in the
    #: botocraft config.
    relations: Dict[str, ModelRelationshipDefinition] = {}
    #: Methods on the model that call methods on the manager for the model.
    manager_methods: Dict[str, ModelManagerMethodDefinition] = {}
    #: If ``True``, this is a bespoke model.  This means that we
    #: don't have a botocore shape for it and are buiding it out of
    #: many AWS API calls.  This is used for things like S3 buckets
    #: and SQS queues, which have no full shapes in botocore.
    bespoke: bool = False

    def unalias_field_name(self, field_name: str) -> str:
        """
        Get the botocore field name for a field name that may be an alias.

        Args:
            field_name: the field name or alias to look up

        Raises:
            KeyError: no such field name or alias

        Returns:
            The botocore field name

        """
        if field_name in self.fields:
            return field_name
        for botocore_name, field_def in self.fields.items():
            if field_def.rename == field_name:
                return botocore_name
        msg = f'{self.name}: No field or alias named "{field_name}"'
        raise KeyError(msg)


# --------
# Managers
# --------


class MethodArgumentDefinition(BaseModel):
    """
    The definition of a single argument on a :py:class:`ManagerMethodDefinition`.

    Example:

        .. code-block:: yaml
            :emphasize-lines: 6,7

            Repository:
                methods:
                    create:
                        boto3_name: create_repository
                        args:
                            repositoryName:
                                required: true

    """

    #: If specified, when constructing the boto3 call invocation, set the value
    #: of this argument to the value of the model attribute specified
    #: here.  This is only useful for methods that take a model instance
    #: as an argument.
    attribute: str | None = None
    #: If specified, in the method signature, use this as the argument
    #: name.  Otherwise, we'll use the name of the boto3 argument.
    rename: str | None = None
    #: If specified, when constructing the boto3 call invocation, set the value
    #: of this argument to the value of this method argument
    source_arg: str | None = None
    #: If specified, in the boto3 call invocation, set the value
    #: of this argument to this specifically.  If this is set, we hide
    #: the argument from the method signature.
    value: Any | None = None
    #: If ``True``, take value as the raw value to pass to the boto3 call
    #: instead of wrapping it in self.serialize()
    raw_value: bool = False
    #: If ``True``, force the argument to be in in the method
    #: signature.  This is only useful for methods that take a model
    #: instance as an argument.
    explicit: bool = False
    #: If ``True``, make this a postional argument.  If ``False``, make
    #: this a keyword argument.
    required: bool = False
    #: If specified, set this as the default value for the argument.
    default: Any | None = None
    #: If ``True``, don't include this argument in the method signature.
    #: or in the boto3 call invocation.
    hidden: bool = False
    #: If specified, use this as the docstring for the argument.  Otherwise,
    #: we'll use the docstring from the botocore operation.
    docstring: str | None = None
    #: If supplied, use this as the python type for the argument.
    python_type: str | None = None
    #: If supplied, this is a list of import strings to add to the top of the
    #: of the file to support the python type.
    imports: List[str] = Field(default_factory=list)


class ManagerMethodDefinition(BaseModel):
    """
    The definition of a single method on a :py:class:`ManagerDefinition`.

    Example:

        .. code-block:: yaml
            :emphasize-lines: 3,4,5,6,7

            Repository:
                methods:
                    create:
                        boto3_name: create_repository
                        args:
                            repositoryName:
                                required: true

    """

    #: The name of the boto3 operation to call in the method body
    boto3_name: str
    #: A list of argument overrides for this method.  If an argument name
    #: is not specified as a key in this dict, it will be generated as is
    #: appropriate for the operation type
    args: Dict[str, MethodArgumentDefinition] = {}
    #: The attribute name on the response class to use in building our
    #: method return object(s).  If not specified, we'll use the lowercased
    #: model name, pluralizing if necessary.
    #:
    #: If you want to just return the whole response from the boto3 call,
    #: set this to literal ``"None"``.
    #:
    #: .. important::
    #:     If you've renamed the model attribute you want to use on the response class
    #:     using :py:attr:`ModelAttributeDefinition.rename`, put the name of
    #:     the botocore here, not the value of the ``rename`` attribute.
    response_attr: str | None = None
    #: If specified, use this as the docstring for the method itself (not
    #: the args, kwargs or return type).  If not specified, we'll use the
    #: docstring from the botocore operation.
    docstring: str | None = None
    #: If specified, use this as the return type for the method.  If not
    #: set, we'll use the model name, pluralizing if necessary.
    return_type: str | None = None
    #: Extra arguments for the method call
    extra_args: Dict[str, MethodArgumentDefinition] = {}
    #: Decorators to wrap the method in
    decorators: List[Importable] = []

    @property
    def explicit_args(self) -> List[str]:
        """
        Return a list of positional arguments that should be explicitly
        included in the method signature.  This is useful for methods
        that take a model instance as an argument, but some of the
        arguments are not model attributes.

        Returns:
            A list of explicit positional arguments

        """
        if not self.args:
            return []
        positional_args = []
        for arg_name, arg_def in self.args.items():
            if arg_def.hidden:
                continue
            if arg_def.explicit and arg_def.required:
                positional_args.append(arg_name)
        return positional_args

    @property
    def explicit_kwargs(self) -> List[str]:
        """
        Return a list of keyword arguments that should be explicitly
        included in the method signature.  This is useful for methods
        that take a model instance as an argument, but some of the
        arguments are not model attributes.

        Returns:
            A list of explicit keyword arguments

        """
        if not self.args:
            return []
        positional_args = []
        for arg_name, arg_def in self.args.items():
            if arg_def.hidden:
                continue
            if arg_def.explicit and not arg_def.required:
                positional_args.append(arg_name)
        return positional_args


@dataclass
class MethodDocstringDefinition:
    """
    A class to hold the different parts of the docstring for a method, and
    render them into a single sphinx-napoleon style docstring.
    """

    #: The main method docstring.  This is the description of the method.
    method: str | None = None
    #: The docstrings for our positional arguments
    args: OrderedDict[str, str | None] = field(default_factory=OrderedDict)
    #: The docstrings for our keyword arguments
    kwargs: OrderedDict[str, str | None] = field(default_factory=OrderedDict)
    #: The docstring for our return value
    return_value: str | None = None

    @property
    def is_empty(self) -> bool:
        """
        Return ``True`` if the docstring is empty, ``False`` otherwise.
        """
        return (
            self.method is None
            and not self.args
            and not self.kwargs
            and self.return_value is None
        )

    def Args(self, formatter: DocumentationFormatter) -> str:
        """
        Return the docstring for the positional arguments.
        """
        assert self.args, "No kwargs"
        docstring = """
        Args:
"""
        for arg_name, arg_docstring in self.args.items():
            docstring += formatter.format_argument(arg_name, arg_docstring)
            docstring += "\n"

        return docstring

    def Keyword_Args(self, formatter: DocumentationFormatter) -> str:
        """
        Return the docstring for the keyword arguments.
        """
        assert self.kwargs, "No args"
        docstring = """
        Keyword Args:
"""
        for arg_name, arg_docstring in self.kwargs.items():
            docstring += formatter.format_argument(arg_name, arg_docstring)
            docstring += "\n"
        return docstring

    def render(self, formatter: DocumentationFormatter) -> str | None:
        """
        Return the entire method docstring.
        """
        if self.is_empty:
            return None

        docstring = '''
        """
'''
        if self.method:
            method_str = formatter.clean(self.method, max_lines=1).strip()
            docstring += indent(method_str, "        ")
            docstring += "\n"
        if self.args:
            docstring += self.Args(formatter)
        if self.kwargs:
            docstring += self.Keyword_Args(formatter)
        docstring += '\n        """'
        return docstring


class ManagerDefinition(BaseModel):
    """
    The definition of a single manager on a :py:class:`ServiceDefinition`.

    Example:
        .. code-block:: yaml

            Repository:
                methods:
                    create:
                        boto3_name: create_repository
                        args:
                            repositoryName:
                                required: true
                    delete:
                        boto3_name: delete_repository
                        args:
                            repositoryName:
                                required: true
                            force:
                                default: false

    """

    #: The name of the model this manager is for
    name: str
    #: The methods to generate on this manager
    methods: Dict[str, ManagerMethodDefinition] = {}
    #: If ``True``, make this manager use the :py:class:`ReadonlyBoto3ModelManager`
    #: superclass
    readonly: bool = False
    #: Mixin classes to add to the manager
    mixins: List[Importable] = []


# --------
# Services
# --------


class ServiceDefinition(BaseModel):
    """
    Our definition of a single AWS service, e.g. ``ecs``, ``ec2``, etc.
    """

    #: The global botocraft interface model
    interface: "BotocraftInterface"
    #: The name of the botocore service
    name: str
    #: The models to generate for this service.  These are specifically models
    #: that have managers.
    primary_models: Dict[str, ModelDefinition] = {}
    #: These are models that are used as attributes on other models, or as
    #: response or request classes.  These models don't have managers, but
    #: we sometimes define them to override the default botocore model
    #: configuration, making them readonly or to rename them, for instance.
    secondary_models: Dict[str, ModelDefinition] = {}
    #: The managers to generate for this service
    managers: Dict[str, ManagerDefinition] = {}

    @property
    def models(self) -> Dict[str, ModelDefinition]:
        """
        Return all of the models we've specifically named in the definition for
        this service, both primary and secondary.

        .. important::

            This does not include models that we generate on the fly, like
            dependent models for primary models, or models for the request and
            response classes for methods on managers.
        """
        return {**self.primary_models, **self.secondary_models}

    @classmethod
    def load(cls, name: str, interface: "BotocraftInterface") -> "ServiceDefinition":
        """
        Load a service definition from its YAML files.

        Args:
            name: The name of the AWS service to load
            interface: The global botocraft interface model

        Returns:
            The loaded service definition

        """
        managers: Dict[str, ManagerDefinition] | None = None
        model_path = DATA_DIR / name / "models.yml"
        manager_path = DATA_DIR / name / "managers.yml"
        with model_path.open(encoding="utf-8") as f:
            models = yaml.safe_load(f)
        if manager_path.exists():
            with manager_path.open(encoding="utf-8") as f:
                managers = {
                    name: ManagerDefinition(name=name, **defn)
                    for name, defn in yaml.safe_load(f).items()
                }
        return cls(
            name=name,
            primary_models={
                name: ModelDefinition(name=name, **defn)
                for name, defn in models.get("primary", {}).items()
            },
            secondary_models={
                name: ModelDefinition(name=name, **defn)
                for name, defn in models.get("secondary", {}).items()
            },
            managers=managers,
            interface=interface,
        )

    def resolve_model_name(self, model_name: str) -> str:
        """
        Resolve a model name to its botocore shape name.  We need to
        do this because we can define :py:attr:`ModelDefinition.alternate_name`
        to change the botocraft name for a model.

        Args:
            model_name: The botocraft model name

        Raises:
            KeyError: If the model name is not defined in this service

        Returns:
            The botocore shape name

        """
        known_models = {**self.primary_models, **self.secondary_models}
        if model_name in known_models:
            return model_name
        for name, model in known_models.items():
            if model.alternate_name == model_name:
                return name
        msg = f'Unknown model name "{model_name}"'
        raise KeyError(msg)

    @property
    def safe_service_name(self) -> str:
        """
        Return a safe version of the service name.  This is used to
        generate the import path for the service.

        Returns:
            The safe service name

        """
        return self.name.replace("-", "_")


# ------
# Global
# ------


class BotocraftInterface(BaseModel):
    #: The services to generate
    services: Dict[str, ServiceDefinition] = {}
    #: The model name to import path mappings
    models: Dict[str, str] = {
        # Tag is defined in so many services that we manually
        # defined it in the botocraft.models.tagging module and
        # define it here so it doesn't get generated.
        "Tag": "botocraft.services.common",
        "Filter": "botocraft.services.common",
    }

    def load(self) -> None:
        """
        Load all known service definitions from their YAML files.

        Returns:
            The loaded interface.

        """
        for path in DATA_DIR.iterdir():
            service = path.name
            if path.is_dir():
                self.services[service] = ServiceDefinition.load(service, interface=self)
                for model_name, model in self.services[service].primary_models.items():
                    # Register the model name to import path mapping so that we can
                    # have them available when we're building our inter-model
                    # relationships.
                    _model_name = model.alternate_name or model_name
                    if _model_name in self.models:
                        msg = f'Model {_model_name} already defined "'
                        f'in "{self.models[_model_name]}"'
                        raise ValueError(msg)
                    self.models[_model_name] = (
                        f"botocraft.services.{self.services[service].safe_service_name}"
                    )
                for manager in self.services[service].managers:
                    # Register the manager name to import path mapping so that we can
                    # have them available when we're building our inter-model
                    # relationships.
                    _manager = manager
                    if manager in self.services[service].primary_models:
                        # We already registered this model during :py:meth:`load`
                        model = self.services[service].primary_models[manager]
                        if model.alternate_name:
                            _manager = model.alternate_name
                    _manager = f"{_manager}Manager"
                    self.models[_manager] = (
                        f"botocraft.services.{self.services[service].safe_service_name}"
                    )

    def add_model(self, name: str, service: str) -> None:
        """
        Save a model name to import path mapping.

        Args:
            name: the model name
            service: the AWS service name

        """
        if name in self.models:
            if (
                self.models[name]
                == f"botocraft.services.{self.services[service].safe_service_name}"
            ):
                # We already registered this model during :py:meth:`load`
                return
            msg = f'Model {name} already defined in "{self.models[name]}"'
            raise ValueError(msg)
        self.models[name] = (
            f"botocraft.services.{self.services[service].safe_service_name}"
        )

    def populate_init_py(self):
        """
        Populate the ``__init__.py`` file in the ``botocraft.models``
        package with import statements for all of our models.
        """
        init_path = SERVICES_DIR / "__init__.py"
        with init_path.open("w", encoding="utf-8") as f:
            for service in self.services.values():
                f.write(
                    f"from .{service.safe_service_name} import *  # noqa: F401,F403\n"
                )

    def populate_services_toc(self):
        """
        Populate the ``_services_index.rst`` file in the ``doc/source``
        folder with the ``.. toctree::`` directive for all of our services.
        """
        index_path = DOCS_DIR / "_services_index.rst"
        code = """
.. toctree::
   :caption: AWS Services
   :hidden:

"""
        for service in sorted(self.services):
            code += f"   api/services/{self.services[service].safe_service_name}\n"
        with index_path.open("w", encoding="utf-8") as f:
            f.write(code)

    def populate_services_list(self):
        """
        Populate the ``_services_list.rst`` file in the ``doc/source``
        folder with links to all of our services.
        """
        index_path = DOCS_DIR / "overview" / "_services_list.rst"
        code = ""
        for service in sorted(self.services):
            code += (
                f"- :doc:`/api/services/{self.services[service].safe_service_name}`\n"
            )
        with index_path.open("w", encoding="utf-8") as f:
            f.write(code)

    def update_init_py(self, service: str) -> None:
        """
        Update the ``__init__.py`` file in the ``botocraft.models``
        package with import statements for the given service.

        Args:
            service: The name of the service to update the ``__init__.py``
                file for.

        """
        init_path = SERVICES_DIR / "__init__.py"
        import_line = (
            f"from .{self.services[service].safe_service_name} import *  "
            "# noqa: F401,F403"
        )
        with init_path.open(encoding="utf-8") as f:
            contents = f.read()
        if import_line not in contents:
            with init_path.open("a", encoding="utf-8") as f:
                f.write(f"{import_line}\n")

    def purge_docs(self) -> None:
        """
        Purge all the previously generated documentation.
        """
        path = DOCS_DIR / "api" / "services"
        for service in self.services.values():
            service_path = path / f"{service.safe_service_name}.rst"
            if service_path.exists():
                service_path.unlink()
        path = DOCS_DIR / "_services_index.rst"
        if path.exists():
            path.unlink()
        path = DOCS_DIR / "overview" / "_services_list.rst"
        if path.exists():
            path.unlink()

    def generate(self, service: str | None = None) -> None:
        """
        Generate all our service interfaces.

        Keyword Args:
            service: If specified, only generate the interface for the given
                service.  Otherwise, generate all interfaces.

        """
        from .service import ServiceGenerator

        # First purge all the previously generated documentation
        self.purge_docs()
        if service:
            assert service in self.services, (
                f'No service definition for AWS Service "{service}"'
            )
            print(f"Generating {service} service interface")
            generator = ServiceGenerator(self.services[service])
            generator.generate()
            self.update_init_py(service)
        else:
            for service_name, _service in self.services.items():
                print(f"Generating {service_name} service interface")
                generator = ServiceGenerator(_service)
                generator.generate()
            self.populate_init_py()
            self.populate_services_toc()
            self.populate_services_list()



# This file is automatically generated by botocraft.  Do not edit directly.
# mypy: disable-error-code="index, override, assignment, union-attr, misc"
from botocraft.mixins.events import EventRule_purge_CreatedBy_attribute
from botocraft.mixins.events import DescribeRuleResponse_to_EventRule
from botocraft.mixins.tags import TagsDictMixin
from .abstract import PrimaryBoto3ModelQuerySet
from collections import OrderedDict
from botocraft.mixins.events import event_rules_only
from typing import ClassVar, Type, Optional, Literal, Dict, List, Union, Any, cast
from datetime import datetime
from botocraft.services.common import Tag
from pydantic import Field
from botocraft.mixins.events import DescribeEventBusResponse_to_EventBus
from .abstract import (
    Boto3Model,
    ReadonlyBoto3Model,
    PrimaryBoto3Model,
    ReadonlyPrimaryBoto3Model,
    Boto3ModelManager,
    ReadonlyBoto3ModelManager,
)
from functools import cached_property

# ===============
# Managers
# ===============


class EventRuleManager(Boto3ModelManager):
    service_name: str = "events"

    @EventRule_purge_CreatedBy_attribute
    def create(self, model: "EventRule", Tags: List[Tag] | None = None) -> str:
        """
        Creates or updates the specified rule. Rules are enabled by default, or based on value of the state. You can disable a
        rule using `DisableRule <https://docs.aws.amazon.com/eventbridge/latest/APIReference/API_DisableRule.html>`_.

        Args:
            model: The :py:class:`Rule` to create.

        Keyword Args:
            Tags: The list of key-value pairs to associate with the rule.

        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            ScheduleExpression=data.get("ScheduleExpression"),
            EventPattern=data.get("EventPattern"),
            State=data.get("State"),
            Description=data.get("Description"),
            RoleArn=data.get("RoleArn"),
            Tags=self.serialize(Tags),
            EventBusName=data.get("EventBusName"),
        )
        _response = self.client.put_rule(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutRuleResponse(**_response)

        self.sessionize(response.RuleArn)
        return cast("str", response.RuleArn)

    @EventRule_purge_CreatedBy_attribute
    def update(self, model: "EventRule", Tags: List[Tag] | None = None) -> str:
        """
        Creates or updates the specified rule. Rules are enabled by default, or based on value of the state. You can disable a
        rule using `DisableRule <https://docs.aws.amazon.com/eventbridge/latest/APIReference/API_DisableRule.html>`_.

        Args:
            model: The :py:class:`Rule` to update.

        Keyword Args:
            Tags: The list of key-value pairs to associate with the rule.

        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            ScheduleExpression=data.get("ScheduleExpression"),
            EventPattern=data.get("EventPattern"),
            State=data.get("State"),
            Description=data.get("Description"),
            RoleArn=data.get("RoleArn"),
            Tags=self.serialize(Tags),
            EventBusName=data.get("EventBusName"),
        )
        _response = self.client.put_rule(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutRuleResponse(**_response)

        self.sessionize(response.RuleArn)
        return cast("str", response.RuleArn)

    def delete(
        self, Name: str, *, EventBusName: str | None = None, Force: bool = True
    ) -> None:
        """
        Deletes the specified rule.

        Args:
            Name: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
            Force: If this is a managed rule, created by an Amazon Web Services service on your behalf, you must specify
                ``Force`` as ``True`` to delete the rule. This parameter is ignored for rules that are not managed rules. You can
                check whether a rule is a managed rule by using ``DescribeRule`` or ``ListRules`` and checking the ``ManagedBy``
                field of the response.
        """
        args: Dict[str, Any] = dict(
            Name=self.serialize(Name),
            EventBusName=self.serialize(EventBusName),
            Force=self.serialize(Force),
        )
        self.client.delete_rule(**{k: v for k, v in args.items() if v is not None})

    @DescribeRuleResponse_to_EventRule
    def get(
        self, Name: str, *, EventBusName: str | None = None
    ) -> Optional["DescribeRuleResponse"]:
        """
        Describes the specified rule.

        Args:
            Name: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
        """
        args: Dict[str, Any] = dict(
            Name=self.serialize(Name), EventBusName=self.serialize(EventBusName)
        )
        _response = self.client.describe_rule(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeRuleResponse(**_response)

        if response:
            self.sessionize(response)
            return response
        return None

    def list(
        self,
        *,
        NamePrefix: str | None = None,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Lists your Amazon EventBridge rules. You can either list all the rules or you
        can provide a prefix to match to the rule names.

        Keyword Args:
            NamePrefix: The prefix matching the rule name.
            EventBusName: The name or ARN of the event bus to list the rules for. If you omit this, the default event bus is
                used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_rules")
        args: Dict[str, Any] = dict(
            NamePrefix=self.serialize(NamePrefix),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = ListRulesResponse(**_response)
            if response.Rules:
                results.extend(response.Rules)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    @event_rules_only
    def list_by_target(
        self,
        TargetArn: str,
        *,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> List[str]:
        """
        Lists the rules for the specified target. You can see which of the rules in
        Amazon EventBridge can invoke a specific target in your account.

        Args:
            TargetArn: The Amazon Resource Name (ARN) of the target resource.

        Keyword Args:
            EventBusName: The name or ARN of the event bus to list rules for. If you omit this, the default event bus is used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_rule_names_by_target")
        args: Dict[str, Any] = dict(
            TargetArn=self.serialize(TargetArn),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List[str] = []

        for _response in response_iterator:
            response = ListRuleNamesByTargetResponse(**_response)
            if response.RuleNames is not None:
                results.extend(response.RuleNames)
            else:
                break

        self.sessionize(results)
        return cast("List[str]", results)

    def list_targets(
        self,
        Rule: str,
        *,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> List["EventTarget"]:
        """
        Lists the targets assigned to the specified rule.

        Args:
            Rule: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_targets_by_rule")
        args: Dict[str, Any] = dict(
            Rule=self.serialize(Rule),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List["EventTarget"] = []

        for _response in response_iterator:
            response = ListTargetsByRuleResponse(**_response)
            if response.Targets is not None:
                results.extend(response.Targets)
            else:
                break

        self.sessionize(results)
        return cast("List[EventTarget]", results)

    def enable(self, Name: str, *, EventBusName: str | None = None) -> None:
        """
        Enables the specified rule. If the rule does not exist, the operation fails.

        Args:
            Name: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
        """
        args: Dict[str, Any] = dict(
            Name=self.serialize(Name), EventBusName=self.serialize(EventBusName)
        )
        self.client.enable_rule(**{k: v for k, v in args.items() if v is not None})

    def disable(self, Name: str, *, EventBusName: str | None = None) -> None:
        """
        Disables the specified rule. A disabled rule won't match any events, and won't
        self-trigger if it has a schedule expression.

        Args:
            Name: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
        """
        args: Dict[str, Any] = dict(
            Name=self.serialize(Name), EventBusName=self.serialize(EventBusName)
        )
        self.client.disable_rule(**{k: v for k, v in args.items() if v is not None})


class EventTargetManager(Boto3ModelManager):
    service_name: str = "events"

    def create(
        self, model: "EventTarget", EventBusName: str | None = None
    ) -> "PutTargetsResponse":
        """
        Adds the specified targets to the specified rule, or updates the targets if they
        are already associated with the rule.

        Args:
            model: The :py:class:`Target` to create.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Rule=data.get("Rule"),
            Targets=data.get("Targets"),
            EventBusName=self.serialize(EventBusName),
        )
        _response = self.client.put_targets(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutTargetsResponse(**_response)

        self.sessionize(response)
        return cast("PutTargetsResponse", response)

    def update(
        self, model: "EventTarget", EventBusName: str | None = None
    ) -> "PutTargetsResponse":
        """
        Adds the specified targets to the specified rule, or updates the targets if they
        are already associated with the rule.

        Args:
            model: The :py:class:`Target` to update.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Rule=data.get("Rule"),
            Targets=data.get("Targets"),
            EventBusName=self.serialize(EventBusName),
        )
        _response = self.client.put_targets(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutTargetsResponse(**_response)

        self.sessionize(response)
        return cast("PutTargetsResponse", response)

    def delete(
        self,
        Rule: str,
        Ids: List[str],
        *,
        EventBusName: str | None = None,
        Force: bool | None = None,
    ) -> "EventTarget":
        """
        Removes the specified targets from the specified rule. When the rule is
        triggered, those targets are no longer be invoked.

        Args:
            Rule: The name of the rule.
            Ids: The IDs of the targets to remove from the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
            Force: If this is a managed rule, created by an Amazon Web Services service on your behalf, you must specify
                ``Force`` as ``True`` to remove targets. This parameter is ignored for rules that are not managed rules. You can
                check whether a rule is a managed rule by using ``DescribeRule`` or ``ListRules`` and checking the ``ManagedBy``
                field of the response.
        """
        args: Dict[str, Any] = dict(
            Rule=self.serialize(Rule),
            Ids=self.serialize(Ids),
            EventBusName=self.serialize(EventBusName),
            Force=self.serialize(Force),
        )
        _response = self.client.remove_targets(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = RemoveTargetsResponse(**_response)
        return response

    def list(
        self,
        Rule: str,
        *,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Lists the targets assigned to the specified rule.

        Args:
            Rule: The name of the rule.

        Keyword Args:
            EventBusName: The name or ARN of the event bus associated with the rule. If you omit this, the default event bus is
                used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_targets_by_rule")
        args: Dict[str, Any] = dict(
            Rule=self.serialize(Rule),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ["ResponseMetadata"]:
                break
            if "ResponseMetadata" in _response:
                del _response["ResponseMetadata"]
            response = ListTargetsByRuleResponse(**_response)
            if response.Targets:
                results.extend(response.Targets)
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        return results

    @event_rules_only
    def list_rules(
        self,
        TargetArn: str,
        *,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> List[str]:
        """
        Lists the rules for the specified target. You can see which of the rules in
        Amazon EventBridge can invoke a specific target in your account.

        Args:
            TargetArn: The Amazon Resource Name (ARN) of the target resource.

        Keyword Args:
            EventBusName: The name or ARN of the event bus to list rules for. If you omit this, the default event bus is used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_rule_names_by_target")
        args: Dict[str, Any] = dict(
            TargetArn=self.serialize(TargetArn),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List[str] = []

        for _response in response_iterator:
            response = ListRuleNamesByTargetResponse(**_response)
            if response.RuleNames is not None:
                results.extend(response.RuleNames)
            else:
                break

        self.sessionize(results)
        return cast("List[str]", results)


class EventBusManager(Boto3ModelManager):
    service_name: str = "events"

    def create(
        self,
        model: "EventBus",
        EventSourceName: str | None = None,
        KmsKeyIdentifier: str | None = None,
        DeadLetterConfig: Optional["EventsDeadLetterConfig"] = None,
        LogConfig: Optional["EventsLogConfig"] = None,
        Tags: List[Tag] | None = None,
    ) -> "CreateEventBusResponse":
        """
        Creates a new event bus within your account. This can be a custom event bus
        which you can use to receive events from your custom applications and services,
        or it can be a partner event bus which can be matched to a partner event source.

        Args:
            model: The :py:class:`EventBus` to create.

        Keyword Args:
            EventSourceName: If you are creating a partner event bus, this specifies the partner event source that the new event
                bus will be matched with.
            KmsKeyIdentifier: The identifier of the KMS customer managed key for EventBridge to use, if you choose to use a
                customer managed key to encrypt events on this event bus. The identifier can be the key Amazon Resource Name (ARN),
                KeyId, key alias, or key alias ARN.
            DeadLetterConfig: Configuration details of the Amazon SQS queue for EventBridge to use as a dead-letter queue (DLQ).
            LogConfig: The logging configuration settings for the event bus.
            Tags: Tags to associate with the event bus.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            EventSourceName=self.serialize(EventSourceName),
            Description=data.get("Description"),
            KmsKeyIdentifier=self.serialize(KmsKeyIdentifier),
            DeadLetterConfig=self.serialize(DeadLetterConfig),
            LogConfig=self.serialize(LogConfig),
            Tags=self.serialize(Tags),
        )
        _response = self.client.create_event_bus(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = CreateEventBusResponse(**_response)

        self.sessionize(response)
        return cast("CreateEventBusResponse", response)

    def update(
        self,
        model: "EventBus",
        KmsKeyIdentifier: str | None = None,
        DeadLetterConfig: Optional["EventsDeadLetterConfig"] = None,
        LogConfig: Optional["EventsLogConfig"] = None,
    ) -> "UpdateEventBusResponse":
        """
        Updates the specified event bus.

        Args:
            model: The :py:class:`EventBus` to update.

        Keyword Args:
            KmsKeyIdentifier: The identifier of the KMS customer managed key for EventBridge to use, if you choose to use a
                customer managed key to encrypt events on this event bus. The identifier can be the key Amazon Resource Name (ARN),
                KeyId, key alias, or key alias ARN.
            DeadLetterConfig: Configuration details of the Amazon SQS queue for EventBridge to use as a dead-letter queue (DLQ).
            LogConfig: The logging configuration settings for the event bus.
        """
        data = model.model_dump(exclude_none=True, by_alias=True)
        args = dict(
            Name=data.get("Name"),
            KmsKeyIdentifier=self.serialize(KmsKeyIdentifier),
            Description=data.get("Description"),
            DeadLetterConfig=self.serialize(DeadLetterConfig),
            LogConfig=self.serialize(LogConfig),
        )
        _response = self.client.update_event_bus(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = UpdateEventBusResponse(**_response)

        self.sessionize(response)
        return cast("UpdateEventBusResponse", response)

    def delete(self, Name: str) -> None:
        """
        Deletes the specified custom event bus or partner event bus. All rules
        associated with this event bus need to be deleted. You can't delete your
        account's default event bus.

        Args:
            Name: The name of the event bus to delete.
        """
        args: Dict[str, Any] = dict(Name=self.serialize(Name))
        self.client.delete_event_bus(**{k: v for k, v in args.items() if v is not None})

    @DescribeEventBusResponse_to_EventBus
    def get(self, *, Name: str | None = None) -> Optional["EventBus"]:
        """
        Displays details about an event bus in your account. This can include the
        external Amazon Web Services accounts that are permitted to write events to your
        default event bus, and the associated policy. For custom event buses and partner
        event buses, it displays the name, ARN, policy, state, and creation time.

        Keyword Args:
            Name: The name or ARN of the event bus to show details for. If you omit this, the default event bus is displayed.
        """
        args: Dict[str, Any] = dict(Name=self.serialize(Name))
        _response = self.client.describe_event_bus(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = DescribeEventBusResponse(**_response)

        if response:
            self.sessionize(response)
            return response
        return None

    def list(
        self,
        *,
        NamePrefix: str | None = None,
        NextToken: str | None = None,
        Limit: int | None = None,
    ) -> PrimaryBoto3ModelQuerySet:
        """
        Lists all the event buses in your account, including the default event bus,
        custom event buses, and partner event buses.

        Keyword Args:
            NamePrefix: Specifying this limits the results to only those event buses with names that start with the specified
                prefix.
            NextToken: The token returned by a previous call, which you can use to retrieve the next set of results.
            Limit: Specifying this limits the number of results returned by this operation. The operation also returns a
                NextToken which you can use in a subsequent operation to retrieve the next set of results.
        """
        args: Dict[str, Any] = dict(
            NamePrefix=self.serialize(NamePrefix),
            NextToken=self.serialize(NextToken),
            Limit=self.serialize(Limit),
        )
        _response = self.client.list_event_buses(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = ListEventBusesResponse(**_response)
        if response and response.EventBuses:
            self.sessionize(response.EventBuses)
            return PrimaryBoto3ModelQuerySet(response.EventBuses)
        return PrimaryBoto3ModelQuerySet([])

    def put_events(
        self,
        Entries: List["PutEventsRequestEntry"],
        *,
        EndpointId: str | None = None,
    ) -> "PutEventsResponse":
        """
        Sends custom events to Amazon EventBridge so that they can be matched to rules.

        Args:
            Entries: The entry that defines an event in your system. You can specify several parameters for the entry such as
                the source and type of the event, resources associated with the event, and so on.

        Keyword Args:
            EndpointId: The URL subdomain of the endpoint. For example, if the URL for Endpoint is
                https://abcde.veo.endpoints.event.amazonaws.com, then the EndpointId is ``abcde.veo``.
        """
        args: Dict[str, Any] = dict(
            Entries=self.serialize(Entries), EndpointId=self.serialize(EndpointId)
        )
        _response = self.client.put_events(
            **{k: v for k, v in args.items() if v is not None}
        )
        response = PutEventsResponse(**_response)

        results: "PutEventsResponse" = None
        if response is not None:
            results = response

        self.sessionize(results)
        return cast("PutEventsResponse", results)

    def list_rules(
        self,
        *,
        NamePrefix: str | None = None,
        EventBusName: str | None = None,
        Limit: int | None = None,
    ) -> List["EventRule"]:
        """
        Lists your Amazon EventBridge rules. You can either list all the rules or you
        can provide a prefix to match to the rule names.

        Keyword Args:
            NamePrefix: The prefix matching the rule name.
            EventBusName: The name or ARN of the event bus to list the rules for. If you omit this, the default event bus is
                used.
            Limit: The maximum number of results to return.
        """
        paginator = self.client.get_paginator("list_rules")
        args: Dict[str, Any] = dict(
            NamePrefix=self.serialize(NamePrefix),
            EventBusName=self.serialize(EventBusName),
            Limit=self.serialize(Limit),
        )
        response_iterator = paginator.paginate(
            **{k: v for k, v in args.items() if v is not None}
        )

        results: List["EventRule"] = []

        for _response in response_iterator:
            response = ListRulesResponse(**_response)
            if response.Rules is not None:
                results.extend(response.Rules)
            else:
                break

        self.sessionize(results)
        return cast("List[EventRule]", results)


# ==============
# Service Models
# ==============


class EventRule(PrimaryBoto3Model):
    """
    Contains information about a rule in Amazon EventBridge.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = EventRuleManager

    Name: str | None = None
    """
    The name of the rule.
    """
    Arn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the rule.
    """
    EventPattern: str | None = None
    """
    The event pattern of the rule.

    For more information, see
    `Events and Event Patterns <https://docs.aws.amazon.com/eventbridge/latest/userguide/eventbridge-and-event-patterns.html>`_
    in the  **Amazon
    EventBridge User Guide** .
    """
    State: Literal["ENABLED", "DISABLED", "ENABLED_WITH_ALL_CLOUDTRAIL_MANAGEMENT_EVENTS"] | None = None
    """
    The state of the rule.
    """
    Description: str | None = None
    """
    The description of the rule.
    """
    ScheduleExpression: str | None = None
    """
    The scheduling expression.

    For example, "cron(0 20 * * ? *)", "rate(5 minutes)". For more information, see
    `Creating an Amazon EventBridge rule that runs on a schedule <https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-create-
    rule-schedule.html>`_.
    """
    RoleArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the role that is used for target invocation.
    """
    ManagedBy: str = Field(default=None, frozen=True)
    """
    If the rule was created on behalf of your account by an Amazon Web Services service,
    this field displays the principal name of the service that created the rule.
    """
    EventBusName: str | None = None
    """
    The name or ARN of the event bus associated with the rule.

    If you omit this, the default event bus is used.
    """
    CreatedBy: str | None = None

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the :py:attr:`Name`
        attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Name

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`Arn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.Arn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`Name`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.Name

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`Name` attribute.
        """
        return hash(self.Name)

    @cached_property
    def targets(self) -> List["EventTarget"] | None:
        """
        Return the targets that are associated with this rule, if any.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "Rule": self.Name,
                }
            )
        except AttributeError:
            return []
        return EventTarget.objects.using(self.session).list(**pk)  # type: ignore[arg-type]

    def enable(self) -> "None":
        """
        Enable the rule.
        """

        return (
            cast("EventRuleManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .enable(
                cast("str", self.Name),
            )
        )

    def disable(self) -> "None":
        """
        Disable the rule.
        """

        return (
            cast("EventRuleManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .disable(
                cast("str", self.Name),
            )
        )


class EventsInputTransformer(Boto3Model):
    """
    Contains the parameters needed for you to provide custom input to a target based on
    one or more pieces of data extracted from the event.
    """

    InputPathsMap: Dict[str, str] | None = None
    """
    Map of JSON paths to be extracted from the event.

    You can then insert these in the template in ``InputTemplate`` to
    produce the output you want to be sent to the target.
    """
    InputTemplate: str
    """
    Input template where you specify placeholders that will be filled with the values of
    the keys from ``InputPathsMap`` to customize the data sent to the target.

    Enclose each ``InputPathsMaps`` value in brackets: <*value*>
    """


class EventsKinesisParameters(Boto3Model):
    """
    This object enables you to specify a JSON path to extract from the event and use as
    the partition key for the Amazon Kinesis data stream, so that you can control the
    shard to which the event goes.

    If you do not include this parameter,
    the default is to use the ``eventId`` as the partition key.
    """

    PartitionKeyPath: str
    """
    The JSON path to be extracted from the event and used as the partition key.

    For more information, see
    `Amazon Kinesis Streams Key Concepts <https://docs.aws.amazon.com/streams/latest/dev/key-concepts.html#partition-key>`_
    in the *Amazon
    Kinesis Streams Developer Guide*.
    """


class RunCommandTarget(Boto3Model):
    """
    Information about the EC2 instances that are to be sent the command, specified as
    key-value pairs.

    Each
    ``RunCommandTarget`` block can include only one key, but this key may specify multiple values.
    """

    Key: str
    """
    Can be either ``tag:`` *tag-key* or ``InstanceIds``.
    """
    Values: List[str]
    """
    If ``Key`` is ``tag:`` *tag-key*, ``Values`` is a list of tag values.

    If ``Key`` is ``InstanceIds``, ``Values`` is a
    list of Amazon EC2 instance IDs.
    """


class EventsRunCommandParameters(Boto3Model):
    """
    This parameter contains the criteria (either InstanceIds or a tag) used to specify
    which EC2 instances are to be sent the command.
    """

    RunCommandTargets: List["RunCommandTarget"]
    """
    Currently, we support including only one RunCommandTarget block, which specifies
    either an array of InstanceIds or a tag.
    """


class EventsAwsVpcConfiguration(Boto3Model):
    """
    This structure specifies the VPC subnets and security groups for the task, and
    whether a public IP address is to be used.

    This structure is relevant only for ECS tasks that use the ``awsvpc`` network mode.
    """

    Subnets: List[str]
    """
    Specifies the subnets associated with the task.

    These subnets must all be in the same VPC. You can specify as many as 16 subnets.
    """
    SecurityGroups: List[str] | None = None
    """
    Specifies the security groups associated with the task.

    These security groups must all be in the same VPC. You can specify as many as five
    security groups. If you do not specify a security group, the default security group
    for the VPC is used.
    """
    AssignPublicIp: Literal["ENABLED", "DISABLED"] | None = None
    """
    Specifies whether the task's elastic network interface receives a public IP address.

    You can specify ``ENABLED`` only
    when ``LaunchType`` in ``EcsParameters`` is set to ``FARGATE``.
    """


class EventsNetworkConfiguration(Boto3Model):
    """
    This structure specifies the network configuration for an ECS task.
    """

    awsvpcConfiguration: EventsAwsVpcConfiguration | None = None
    """
    Use this structure to specify the VPC subnets and security groups for the task, and
    whether a public IP address is to be used.

    This structure is relevant only for ECS tasks that use the ``awsvpc`` network mode.
    """


class EventsCapacityProviderStrategyItem(Boto3Model):
    """
    The details of a capacity provider strategy.

    To learn more, see `CapacityProviderStrategyItem <https://docs.aws.amazon.c
    om/AmazonECS/latest/APIReference/API_CapacityProviderStrategyItem.html>`_ in the Amazon ECS API Reference.
    """

    capacityProvider: str
    """
    The short name of the capacity provider.
    """
    weight: int | None = None
    """
    The weight value designates the relative percentage of the total number of tasks
    launched that should use the specified capacity provider.

    The weight value is taken into consideration after the base value, if defined, is
    satisfied.
    """
    base: int | None = None
    """
    The base value designates how many tasks, at a minimum, to run on the specified
    capacity provider.

    Only one capacity provider in a capacity provider strategy can have a base defined.
    If no value is specified, the default value of 0 is used.
    """


class EventsPlacementConstraint(Boto3Model):
    """
    An object representing a constraint on task placement.

    To learn more, see
    `Task Placement Constraints <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-placement-constraints.html>`_
    in the Amazon
    Elastic Container Service Developer Guide.
    """

    type: Literal["distinctInstance", "memberOf"] | None = None
    """
    The type of constraint.

    Use distinctInstance to ensure that each task in a particular group is running on a
    different container instance. Use memberOf to restrict the selection to a group of
    valid candidates.
    """
    expression: str | None = None
    """
    A cluster query language expression to apply to the constraint.

    You cannot specify an expression if the constraint type
    is ``distinctInstance``. To learn more, see `Cluster Query
    Language <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/cluster-query-language.html>`_ in the Amazon Elastic
    Container Service Developer Guide.
    """


class EventsPlacementStrategy(Boto3Model):
    """
    The task placement strategy for a task or service.

    To learn more, see
    `Task Placement Strategies <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task-placement-strategies.html>`_
    in the Amazon
    Elastic Container Service Service Developer Guide.
    """

    type: Literal["random", "spread", "binpack"] | None = None
    """
    The type of placement strategy.

    The random placement strategy randomly places tasks on available candidates. The
    spread placement strategy spreads placement across available candidates evenly based
    on the field parameter. The binpack strategy places tasks on available candidates
    that have the least available amount of the resource that is specified with the
    field parameter. For example, if you binpack on memory, a task is placed on the
    instance with the least amount of remaining memory (but still enough to run the
    task).
    """
    field: str | None = None
    """
    The field to apply the placement strategy against.

    For the spread placement strategy, valid values are instanceId (or host, which has
    the same effect), or any platform or custom attribute that is applied to a container
    instance, such as attribute:ecs.availability-zone. For the binpack placement
    strategy, valid values are cpu and memory. For the random placement strategy, this
    field is not used.
    """


class EventsEcsParameters(TagsDictMixin, Boto3Model):
    """
    The custom parameters to be used when the target is an Amazon ECS task.
    """

    tag_class: ClassVar[Type] = Tag
    TaskDefinitionArn: str
    """
    The ARN of the task definition to use if the event target is an Amazon ECS task.
    """
    TaskCount: int | None = None
    """
    The number of tasks to create based on ``TaskDefinition``.

    The default is 1.
    """
    LaunchType: Literal["EC2", "FARGATE", "EXTERNAL"] | None = None
    """
    Specifies the launch type on which your task is running.

    The launch type that you specify here must match one of the
    launch type (compatibilities) of the target task. The ``FARGATE`` value is supported only in the Regions where Fargate
    with Amazon ECS is supported. For more information, see `Fargate on Amazon
    ECS <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/AWS-Fargate.html>`_ in the *Amazon Elastic Container
    Service Developer Guide*.
    """
    NetworkConfiguration: EventsNetworkConfiguration | None = None
    """
    Use this structure if the Amazon ECS task uses the ``awsvpc`` network mode.

    This structure specifies the VPC subnets and
    security groups associated with the task, and whether a public IP address is to be used. This structure is required if
    ``LaunchType`` is ``FARGATE`` because the ``awsvpc`` mode is required for Fargate tasks.
    """
    PlatformVersion: str | None = None
    """
    Specifies the platform version for the task.

    Specify only the numeric portion of the platform version, such as
    ``1.1.0``.
    """
    Group: str | None = None
    """
    Specifies an ECS task group for the task.

    The maximum length is 255 characters.
    """
    CapacityProviderStrategy: List["EventsCapacityProviderStrategyItem"] | None = (
        None
    )
    """
    The capacity provider strategy to use for the task.
    """
    EnableECSManagedTags: bool | None = None
    """
    Specifies whether to enable Amazon ECS managed tags for the task.

    For more information, see
    `Tagging Your Amazon ECS Resources <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/ecs-using-tags.html>`_
    in the Amazon Elastic
    Container Service Developer Guide.
    """
    EnableExecuteCommand: bool | None = None
    """
    Whether or not to enable the execute command functionality for the containers in
    this task.

    If true, this enables execute command functionality on all containers in the task.
    """
    PlacementConstraints: List["EventsPlacementConstraint"] | None = None
    """
    An array of placement constraint objects to use for the task.

    You can specify up to 10 constraints per task (including constraints in the task
    definition and those specified at runtime).
    """
    PlacementStrategy: List["EventsPlacementStrategy"] | None = None
    """
    The placement strategy objects to use for the task.

    You can specify a maximum of five strategy rules per task.
    """
    PropagateTags: Literal["TASK_DEFINITION"] | None = None
    """
    Specifies whether to propagate the tags from the task definition to the task.

    If no value is specified, the tags are not propagated. Tags can only be propagated
    to the task during task creation. To add tags to a task after task creation, use the
    TagResource API action.
    """
    ReferenceId: str | None = None
    """
    The reference ID to use for the task.
    """
    Tags: List[Tag] | None = None
    """
    The metadata that you apply to the task to help you categorize and organize them.

    Each tag consists of a key and an optional value, both of which you define. To learn
    more, see
    `RunTask <https://docs.aws.amazon.com/AmazonECS/latest/APIReference/API_RunTask.html#ECS-RunTask-request-tags>`_
    in the
    Amazon ECS API Reference.
    """


class BatchArrayProperties(Boto3Model):
    """
    The array properties for the submitted job, such as the size of the array.

    The array size can be between 2 and 10,000. If you specify array properties for a
    job, it becomes an array job. This parameter is used only if the target is an Batch
    job.
    """

    Size: int | None = None
    """
    The size of the array, if this is an array batch job.

    Valid values are integers between 2 and 10,000.
    """


class BatchRetryStrategy(Boto3Model):
    """
    The retry strategy to use for failed jobs, if the target is an Batch job.

    If you specify a retry strategy here, it overrides the retry strategy defined in the
    job definition.
    """

    Attempts: int | None = None
    """
    The number of times to attempt to retry, if the job fails.

    Valid values are 1-10.
    """


class EventsBatchParameters(Boto3Model):
    """
    The custom parameters to be used when the target is an Batch job.
    """

    JobDefinition: str
    """
    The ARN or name of the job definition to use if the event target is an Batch job.

    This job definition must already exist.
    """
    JobName: str
    """
    The name to use for this execution of the job, if the target is an Batch job.
    """
    ArrayProperties: BatchArrayProperties | None = None
    """
    The array properties for the submitted job, such as the size of the array.

    The array size can be between 2 and 10,000. If you specify array properties for a
    job, it becomes an array job. This parameter is used only if the target is an Batch
    job.
    """
    RetryStrategy: BatchRetryStrategy | None = None
    """
    The retry strategy to use for failed jobs, if the target is an Batch job.

    The retry strategy is the number of times to retry the failed job execution. Valid
    values are 1-10. When you specify a retry strategy here, it overrides the retry
    strategy defined in the job definition.
    """


class EventsSqsParameters(Boto3Model):
    """
    This structure includes the custom parameter to be used when the target is an SQS
    FIFO queue.
    """

    MessageGroupId: str | None = None
    """
    The FIFO message group ID to use as the target.
    """


class EventsHttpParameters(Boto3Model):
    """
    These are custom parameter to be used when the target is an API Gateway APIs or
    EventBridge ApiDestinations.

    In the latter case, these are merged with any InvocationParameters specified on the
    Connection, with any values from the Connection taking precedence.
    """

    PathParameterValues: List[str] | None = None
    """
    The path parameter values to be used to populate API Gateway API or EventBridge
    ApiDestination path wildcards ("*").
    """
    HeaderParameters: Dict[str, str] | None = None
    """
    The headers that need to be sent as part of request invoking the API Gateway API or
    EventBridge ApiDestination.
    """
    QueryStringParameters: Dict[str, str] | None = None
    """
    The query string keys/values that need to be sent as part of request invoking the
    API Gateway API or EventBridge ApiDestination.
    """


class EventsRedshiftDataParameters(Boto3Model):
    """
    These are custom parameters to be used when the target is a Amazon Redshift cluster
    to invoke the Amazon Redshift Data API ExecuteStatement based on EventBridge events.
    """

    SecretManagerArn: str | None = None
    """
    The name or ARN of the secret that enables access to the database.

    Required when authenticating using Amazon Web Services Secrets Manager.
    """
    Database: str
    """
    The name of the database.

    Required when authenticating using temporary credentials.
    """
    DbUser: str | None = None
    """
    The database user name.

    Required when authenticating using temporary credentials.
    """
    Sql: str | None = None
    """
    The SQL statement text to run.
    """
    StatementName: str | None = None
    """
    The name of the SQL statement.

    You can name the SQL statement when you create it to identify the query.
    """
    WithEvent: bool | None = None
    """
    Indicates whether to send an event back to EventBridge after the SQL statement runs.
    """
    Sqls: List[str] | None = None
    """
    One or more SQL statements to run.

    The SQL statements are run as a single transaction. They run serially in the order
    of the array. Subsequent SQL statements don't start until the previous statement in
    the array completes. If any SQL statement fails, then because they are run as one
    transaction, all work is rolled back.
    """


class SageMakerPipelineParameter(Boto3Model):
    """
    Name/Value pair of a parameter to start execution of a SageMaker AI Model Building
    Pipeline.
    """

    Name: str
    """
    Name of parameter to start execution of a SageMaker AI Model Building Pipeline.
    """
    Value: str
    """
    Value of parameter to start execution of a SageMaker AI Model Building Pipeline.
    """


class EventsSageMakerPipelineParameters(Boto3Model):
    """
    These are custom parameters to use when the target is a SageMaker AI Model Building
    Pipeline that starts based on EventBridge events.
    """

    PipelineParameterList: List["SageMakerPipelineParameter"] | None = None
    """
    List of Parameter names and values for SageMaker AI Model Building Pipeline
    execution.
    """


class EventsDeadLetterConfig(Boto3Model):
    """
    Configuration details of the Amazon SQS queue for EventBridge to use as a dead-
    letter queue (DLQ).

    For more information, see `Using dead-letter queues to process undelivered
    events <https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-rule-event-delivery.html#eb-rule-dlq>`_ in the
    *EventBridge User Guide*.
    """

    Arn: str | None = None
    """
    The ARN of the SQS queue specified as the target for the dead-letter queue.
    """


class EventsRetryPolicy(Boto3Model):
    """
    A ``RetryPolicy`` object that includes information about the retry policy settings.
    """

    MaximumRetryAttempts: int | None = None
    """
    The maximum number of retry attempts to make before the request fails.

    Retry attempts continue until either the maximum
    number of attempts is made or until the duration of the ``MaximumEventAgeInSeconds`` is met.
    """
    MaximumEventAgeInSeconds: int | None = None
    """
    The maximum amount of time, in seconds, to continue to make retry attempts.
    """


class EventsAppSyncParameters(Boto3Model):
    """
    Contains the GraphQL operation to be parsed and executed, if the event target is an
    AppSync API.
    """

    GraphQLOperation: str | None = None
    """
    The GraphQL operation; that is, the query, mutation, or subscription to be parsed
    and executed by the GraphQL service.
    """


class EventTarget(PrimaryBoto3Model):
    """Targets are the resources to be invoked when a rule is triggered. For a complete list of services and resources that can
    be set as a target, see `PutTargets <https://docs.aws.amazon.com/eventbridge/latest/APIReference/API_PutTargets.html>`_.

    If you are setting the event bus of another account as the target, and that account granted permission to your account
    through an organization instead of directly by the account ID, then you must specify a ``RoleArn`` with proper
    permissions in the ``Target`` structure. For more information, see `Sending and Receiving Events Between Amazon Web
    Services Accounts <https://docs.aws.amazon.com/eventbridge/latest/userguide/eventbridge-cross-account-event-
    delivery.html>`_ in the *Amazon EventBridge User Guide*.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = EventTargetManager

    Id: str = Field(frozen=True)
    """
    The ID of the target within the specified rule.

    Use this ID to reference the target when updating the rule. We recommend using a
    memorable and unique string.
    """
    Arn: str = Field(frozen=True)
    """
    The Amazon Resource Name (ARN) of the target.
    """
    RoleArn: str = Field(default=None, frozen=True)
    """
    The Amazon Resource Name (ARN) of the IAM role to be used for this target when the
    rule is triggered.

    If one rule triggers multiple targets, you can use a different IAM role for each
    target.
    """
    Input: str = Field(default=None, frozen=True)
    """
    Valid JSON text passed to the target.

    In this case, nothing from the event itself is passed to the target. For more
    information, see `The JavaScript Object Notation (JSON) Data Interchange Format <http://www.rfc-
    editor.org/rfc/rfc7159.txt>`_.
    """
    InputPath: str = Field(default=None, frozen=True)
    """
    The value of the JSONPath that is used for extracting part of the matched event when
    passing it to the target.

    You may use JSON dot notation or bracket notation. For more information about JSON
    paths, see
    `JSONPath <http://goessner.net/articles/JsonPath/>`_.
    """
    InputTransformer: EventsInputTransformer = Field(default=None, frozen=True)
    """
    Settings to enable you to provide custom input to a target based on certain event
    data.

    You can extract one or more key- value pairs from the event and then use that data
    to send customized input to the target.
    """
    KinesisParameters: EventsKinesisParameters = Field(default=None, frozen=True)
    """
    The custom parameter you can use to control the shard assignment, when the target is
    a Kinesis data stream.

    If you do
    not include this parameter, the default is to use the ``eventId`` as the partition key.
    """
    RunCommandParameters: EventsRunCommandParameters = Field(default=None, frozen=True)
    """
    Parameters used when you are using the rule to invoke Amazon EC2 Run Command.
    """
    EcsParameters: EventsEcsParameters = Field(default=None, frozen=True)
    """
    Contains the Amazon ECS task definition and task count to be used, if the event
    target is an Amazon ECS task.

    For more information about Amazon ECS tasks, see
    `Task Definitions <https://docs.aws.amazon.com/AmazonECS/latest/developerguide/task_defintions.html>`_
    in the *Amazon EC2
    Container Service Developer Guide*.
    """
    BatchParameters: EventsBatchParameters = Field(default=None, frozen=True)
    """
    If the event target is an Batch job, this contains the job definition, job name, and
    other parameters.

    For more information, see
    `Jobs <https://docs.aws.amazon.com/batch/latest/userguide/jobs.html>`_
    in the *Batch User Guide*.
    """
    SqsParameters: EventsSqsParameters = Field(default=None, frozen=True)
    """
    Contains the message group ID to use when the target is a FIFO queue.
    """
    HttpParameters: EventsHttpParameters = Field(default=None, frozen=True)
    """
    Contains the HTTP parameters to use when the target is a API Gateway endpoint or
    EventBridge ApiDestination.
    """
    RedshiftDataParameters: EventsRedshiftDataParameters = Field(
        default=None, frozen=True
    )
    """
    Contains the Amazon Redshift Data API parameters to use when the target is a Amazon
    Redshift cluster.
    """
    SageMakerPipelineParameters: EventsSageMakerPipelineParameters = Field(
        default=None, frozen=True
    )
    """
    Contains the SageMaker AI Model Building Pipeline parameters to start execution of a
    SageMaker AI Model Building Pipeline.
    """
    DeadLetterConfig: EventsDeadLetterConfig = Field(default=None, frozen=True)
    """
    The ``DeadLetterConfig`` that defines the target queue to send dead-letter queue
    events to.
    """
    RetryPolicy: EventsRetryPolicy = Field(default=None, frozen=True)
    """
    The retry policy configuration to use for the dead-letter queue.
    """
    AppSyncParameters: EventsAppSyncParameters = Field(default=None, frozen=True)
    """
    Contains the GraphQL operation to be parsed and executed, if the event target is an
    AppSync API.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the :py:attr:`Id`
        attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Id

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`Arn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.Arn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`Id`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.Id

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`Id` attribute.
        """
        return hash(self.Id)

    @cached_property
    def rules(self) -> Optional["EventRule"]:
        """
        Return the :py:class:`Rule` object that this target belongs to.

        .. note::

            The output of this property is cached on the model instance, so
            calling this multiple times will not result in multiple calls to the
            AWS API.   If you need a fresh copy of the data, you can re-get the
            model instance from the manager.
        """

        try:
            pk = OrderedDict(
                {
                    "TargetArn": self.Arn,
                }
            )
        except AttributeError:
            return None
        return EventRule.objects.using(self.session).list_by_target(**pk)  # type: ignore[arg-type]


class PutEventsRequestEntry(Boto3Model):
    """
    Represents an event to be submitted.
    """

    Time: datetime | None = None
    """
The time stamp of the event, per `RFC3339 <https://www.rfc-editor.org/rfc/rfc3339.txt>`_. If no time stamp is provided,
the time stamp of the `PutEvents <https://docs.aws.amazon.com/eventbridge/latest/APIReference/API_PutEvents.html>`_ call
is used.
    """
    Source: str | None = None
    """
    The source of the event.
    """
    Resources: List[str] | None = None
    """
    Amazon Web Services resources, identified by Amazon Resource Name (ARN), which the
    event primarily concerns.

    Any number, including zero, may be present.
    """
    DetailType: str | None = None
    """
    Free-form string, with a maximum of 128 characters, used to decide what fields to
    expect in the event detail.
    """
    Detail: str | None = None
    """
    A valid JSON object.

    There is no other schema imposed. The JSON object may contain fields and nested sub-
    objects.
    """
    EventBusName: str | None = None
    """
    The name or ARN of the event bus to receive the event.

    Only the rules that are associated with this event bus are used to match the event.
    If you omit this, the default event bus is used.
    """
    TraceHeader: str | None = None
    """
    An X-Ray trace header, which is an http header (X-Amzn-Trace-Id) that contains the
    trace-id associated with the event.
    """


class EventBus(PrimaryBoto3Model):
    """
    An event bus receives events from a source, uses rules to evaluate them, applies any
    configured input transformation, and routes them to the appropriate target(s).

    Your account's default event bus receives events from Amazon Web Services services.
    A custom event bus can receive events from your custom applications and services. A
    partner event bus receives events from an event source created by an SaaS partner.
    These events come from the partners services or applications.
    """

    manager_class: ClassVar[Type[Boto3ModelManager]] = EventBusManager

    Name: str | None = None
    """
    The name of the event bus.
    """
    Arn: str = Field(default=None, frozen=True)
    """
    The ARN of the event bus.
    """
    Description: str | None = None
    """
    The event bus description.
    """
    Policy: str = Field(default=None, frozen=True)
    """
    The permissions policy of the event bus, describing which other Amazon Web Services
    accounts can write events to this event bus.
    """
    CreationTime: datetime = Field(default=None, frozen=True)
    """
    The time the event bus was created.
    """
    LastModifiedTime: datetime = Field(default=None, frozen=True)
    """
    The time the event bus was last modified.
    """

    @property
    def pk(self) -> str | None:
        """
        Return the primary key of the model.   This is the value of the :py:attr:`Name`
        attribute.

        Returns:
            The primary key of the model instance.
        """
        return self.Name

    @property
    def arn(self) -> str | None:
        """
        Return the ARN of the model.   This is the value of the :py:attr:`Arn`
        attribute.

        Returns:
            The ARN of the model instance.
        """
        return self.Arn

    @property
    def name(self) -> str | None:
        """
        Return the name of the model.   This is the value of the :py:attr:`Name`
        attribute.

        Returns:
            The name of the model instance.
        """
        return self.Name

    def __hash__(self) -> int:
        """
        Return the hash of the model.

        This is the value of the
        :py:attr:`Name` attribute.
        """
        return hash(self.Name)

    def rules(self) -> List["EventRule"]:
        """
        Return the rules that are associated with this event bus.
        """

        return (
            cast("EventBusManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .list_rules(
                cast("str", self.Name),
            )
        )

    def put_events(
        self, Entries: List["PutEventsRequestEntry"], EndpointId: str | None = None
    ) -> "PutEventsResponse":
        """
        Put events to this event bus.

        Args:
            Entries: The entry that defines an event in your system. You can specify several parameters for the entry such as
                the source and type of the event, resources associated with the event, and so on.

        Keyword Args:
            EndpointId: The URL subdomain of the endpoint. For example, if the URL for Endpoint is
                https://abcde.veo.endpoints.event.amazonaws.com, then the EndpointId is ``abcde.veo``.
        """

        return (
            cast("EventBusManager", self.objects)  # type: ignore[attr-defined]
            .using(self.session)
            .put_events(Entries, EndpointId=EndpointId)
        )


# =======================
# Request/Response Models
# =======================


class PutRuleResponse(Boto3Model):
    RuleArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the rule.
    """


class DescribeRuleResponse(Boto3Model):
    Name: str | None = None
    """
    The name of the rule.
    """
    Arn: str | None = None
    """
    The Amazon Resource Name (ARN) of the rule.
    """
    EventPattern: str | None = None
    """
    The event pattern.

    For more information, see
    `Events and Event Patterns <https://docs.aws.amazon.com/eventbridge/latest/userguide/eventbridge-and-event-patterns.html>`_
    in the  **Amazon
    EventBridge User Guide** .
    """
    ScheduleExpression: str | None = None
    """
    The scheduling expression.

    For example, "cron(0 20 * * ? *)", "rate(5 minutes)".
    """
    State: Literal["ENABLED", "DISABLED", "ENABLED_WITH_ALL_CLOUDTRAIL_MANAGEMENT_EVENTS"] | None = None
    """
    Specifies whether the rule is enabled or disabled.
    """
    Description: str | None = None
    """
    The description of the rule.
    """
    RoleArn: str | None = None
    """
    The Amazon Resource Name (ARN) of the IAM role associated with the rule.
    """
    ManagedBy: str | None = None
    """
    If this is a managed rule, created by an Amazon Web Services service on your behalf,
    this field displays the principal name of the Amazon Web Services service that
    created the rule.
    """
    EventBusName: str | None = None
    """
    The name of the event bus associated with the rule.
    """
    CreatedBy: str | None = None
    """
    The account ID of the user that created the rule.

    If you use ``PutRule`` to put a rule on an event bus in another
    account, the other account is the owner of the rule, and the rule ARN includes the account ID for that account. However,
    the value for ``CreatedBy`` is the account ID as the account that created the rule in the other account.
    """


class ListRulesResponse(Boto3Model):
    Rules: List["EventRule"] | None = None
    """
    The rules that match the specified criteria.
    """
    NextToken: str | None = None
    """
    A token indicating there are more results available.

    If there are no more results, no token is included in the response.
    """


class ListRuleNamesByTargetResponse(Boto3Model):
    RuleNames: List[str] | None = None
    """
    The names of the rules that can invoke the given target.
    """
    NextToken: str | None = None
    """
    A token indicating there are more results available.

    If there are no more results, no token is included in the response.
    """


class ListTargetsByRuleResponse(Boto3Model):
    Targets: List["EventTarget"] | None = None
    """
    The targets assigned to the rule.
    """
    NextToken: str | None = None
    """
    A token indicating there are more results available.

    If there are no more results, no token is included in the response.
    """


class PutTargetsResultEntry(Boto3Model):
    """
    Represents a target that failed to be added to a rule.
    """

    TargetId: str | None = None
    """
    The ID of the target.
    """
    ErrorCode: str | None = None
    """
    The error code that indicates why the target addition failed.

    If the value is ``ConcurrentModificationException``, too
    many requests were made at the same time.
    """
    ErrorMessage: str | None = None
    """
    The error message that explains why the target addition failed.
    """


class PutTargetsResponse(Boto3Model):
    FailedEntryCount: int | None = None
    """
    The number of failed entries.
    """
    FailedEntries: List["PutTargetsResultEntry"] | None = None
    """
    The failed target entries.
    """


class RemoveTargetsResultEntry(Boto3Model):
    """
    Represents a target that failed to be removed from a rule.
    """

    TargetId: str | None = None
    """
    The ID of the target.
    """
    ErrorCode: str | None = None
    """
    The error code that indicates why the target removal failed.

    If the value is ``ConcurrentModificationException``, too
    many requests were made at the same time.
    """
    ErrorMessage: str | None = None
    """
    The error message that explains why the target removal failed.
    """


class RemoveTargetsResponse(Boto3Model):
    FailedEntryCount: int | None = None
    """
    The number of failed entries.
    """
    FailedEntries: List["RemoveTargetsResultEntry"] | None = None
    """
    The failed target entries.
    """


class EventsLogConfig(Boto3Model):
    """
    The logging configuration settings for the event bus.

    For more information, see
    `Configuring logs for event buses <https://docs.aws.amazon.com/eb-event-bus-logs.html>`_
    in the
    *EventBridge User Guide*.
    """

    IncludeDetail: Literal["NONE", "FULL"] | None = None
    """
    Whether EventBridge include detailed event information in the records it generates.

    Detailed data can be useful for troubleshooting and debugging. This information
    includes details of the event itself, as well as target details.
    """
    Level: Literal["OFF", "ERROR", "INFO", "TRACE"] | None = None
    """
    The level of logging detail to include.

    This applies to all log destinations for the event bus.
    """


class CreateEventBusResponse(Boto3Model):
    EventBusArn: str | None = None
    """
    The ARN of the new event bus.
    """
    Description: str | None = None
    """
    The event bus description.
    """
    KmsKeyIdentifier: str | None = None
    """
    The identifier of the KMS customer managed key for EventBridge to use to encrypt
    events on this event bus, if one has been specified.
    """
    DeadLetterConfig: EventsDeadLetterConfig | None = None
    """
    Configuration details of the Amazon SQS queue for EventBridge to use as a dead-
    letter queue (DLQ).
    """
    LogConfig: EventsLogConfig | None = None
    """
    The logging configuration settings for the event bus.
    """


class UpdateEventBusResponse(Boto3Model):
    Arn: str | None = None
    """
    The event bus Amazon Resource Name (ARN).
    """
    Name: str | None = None
    """
    The event bus name.
    """
    KmsKeyIdentifier: str | None = None
    """
    The identifier of the KMS customer managed key for EventBridge to use to encrypt
    events on this event bus, if one has been specified.
    """
    Description: str | None = None
    """
    The event bus description.
    """
    DeadLetterConfig: EventsDeadLetterConfig | None = None
    """
    Configuration details of the Amazon SQS queue for EventBridge to use as a dead-
    letter queue (DLQ).
    """
    LogConfig: EventsLogConfig | None = None
    """
    The logging configuration settings for the event bus.
    """


class DescribeEventBusResponse(Boto3Model):
    Name: str | None = None
    """
    The name of the event bus.

    Currently, this is always ``default``.
    """
    Arn: str | None = None
    """
    The Amazon Resource Name (ARN) of the account permitted to write events to the
    current account.
    """
    Description: str | None = None
    """
    The event bus description.
    """
    KmsKeyIdentifier: str | None = None
    """
    The identifier of the KMS customer managed key for EventBridge to use to encrypt
    events on this event bus, if one has been specified.
    """
    DeadLetterConfig: EventsDeadLetterConfig | None = None
    """
    Configuration details of the Amazon SQS queue for EventBridge to use as a dead-
    letter queue (DLQ).
    """
    Policy: str | None = None
    """
    The policy that enables the external account to send events to your account.
    """
    LogConfig: EventsLogConfig | None = None
    """
    The logging configuration settings for the event bus.
    """
    CreationTime: datetime | None = None
    """
    The time the event bus was created.
    """
    LastModifiedTime: datetime | None = None
    """
    The time the event bus was last modified.
    """


class ListEventBusesResponse(Boto3Model):
    EventBuses: List["EventBus"] | None = None
    """
    This list of event buses.
    """
    NextToken: str | None = None
    """
    A token indicating there are more results available.

    If there are no more results, no token is included in the response.
    """


class PutEventsResultEntry(Boto3Model):
    """
    Represents the results of an event submitted to an event bus.

    If the submission was successful, the entry has the event ID in it. Otherwise, you
    can use the error code and error message to identify the problem with the entry.

    For information about the errors that are common to all actions, see
    `Common Errors <https://docs.aws.amazon.com/eventbridge/latest/APIReference/CommonErrors.html>`_.
    """

    EventId: str | None = None
    """
    The ID of the event.
    """
    ErrorCode: str | None = None
    """
    The error code that indicates why the event submission failed.
    """
    ErrorMessage: str | None = None
    """
    The error message that explains why the event submission failed.
    """


class PutEventsResponse(Boto3Model):
    FailedEntryCount: int | None = None
    """
    The number of failed entries.
    """
    Entries: List["PutEventsResultEntry"] | None = None
    """
    The successfully and unsuccessfully ingested events results.

    If the ingestion was successful, the entry has the event ID in it. Otherwise, you
    can use the error code and error message to identify the problem with the entry.
    """



from collections import OrderedDict
from typing import Literal, cast

from .base import ManagerMethodGenerator


class ListMethodGenerator(ManagerMethodGenerator):
    method_name: str = "list"

    def kwargs(
        self, location: Literal["method", "operation"] = "method"
    ) -> OrderedDict[str, str]:
        """
        Override the kwargs to exclude the pagination arguments if
        the boto3 operation can paginate.
        """
        _args: OrderedDict[str, str] = OrderedDict()
        if self.client.can_paginate(self.boto3_name):
            for _arg, arg_type in super().kwargs(location=location).items():
                if _arg not in self.PAGINATOR_ARGS:
                    _args[_arg] = arg_type
        else:
            for _arg, arg_type in super().kwargs(location=location).items():
                _args[_arg] = arg_type
        return _args

    @property
    def return_type(self) -> str:
        """
        For list methods, we return a list of model instances, not the response
        model, unless it's overriden in our botocraft method config, in which
        case we return that.

        Thus we need to change the return type to a list of the model.

        Returns:
            The name of the return type class.

        """
        # We do this because :py:meth:`response_class` will create the response class
        # if it doesn't exist, and we need that to happen so we can use its attributes
        _ = self.response_class
        if self.method_def.return_type:
            return self.method_def.return_type
        if self.output_shape is not None:
            response_attr_shape = self.output_shape.members[
                cast("str", self.response_attr)
            ]
        return_type = self.shape_converter.convert(response_attr_shape, quote=True)
        _return_type = return_type.lower()
        if _return_type.startswith("list["):
            _return_type = _return_type.replace("list[", "").replace("]", "")
        if _return_type.startswith("dict["):
            _return_type = _return_type.replace("dict[", "").replace("]", "")
            _return_type = _return_type.split(",")[1].strip()
        if _return_type in ("str", "int", "float", "bool", "None"):
            return return_type
        return "PrimaryBoto3ModelQuerySet"

    @property
    def body(self) -> str:
        # This is a hard attribute to guess. Sometimes it's CamelCase, sometimes
        # it's camelCase, sometimes it's snake_case.  We'll just assume it's a
        # lowercase plural of the model name.
        self.imports.add("from .abstract import PrimaryBoto3ModelQuerySet")
        if self.client.can_paginate(self.boto3_name):
            code = f"""
        paginator = self.client.get_paginator('{self.boto3_name}')
        {self.operation_args}
        response_iterator = paginator.paginate(**{{k: v for k, v in args.items() if v is not None}})
        results = []
        for _response in response_iterator:
            if list(_response.keys()) == ['ResponseMetadata']:
                break
            if 'ResponseMetadata' in _response:
                del _response['ResponseMetadata']
            response = {self.response_class}(**_response)
            if response.{self.response_attr}:
                results.extend(response.{self.response_attr})
            else:
                if getattr(response, "NextToken", None):
                    continue
                break
        self.sessionize(results)
        if results and isinstance(results[0], Boto3Model):
            return PrimaryBoto3ModelQuerySet(results)
        else:
            return results
"""  # noqa: E501
        else:
            code = f"""
        {self.operation_args}
        {self.operation_call}
        if response and response.{self.response_attr}:
            self.sessionize(response.{self.response_attr})
            return PrimaryBoto3ModelQuerySet(response.{self.response_attr})
        return PrimaryBoto3ModelQuerySet([])
"""
        return code
